"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.GaslessTransaction = void 0;
const web3_js_1 = require("@solana/web3.js");
const types_1 = require("./types");
const gasless_1 = require("../gasless");
const pow_1 = require("../pow");
const anchor_1 = require("@project-serum/anchor");
const helpers_1 = require("../helpers");
const gasless_program_json_1 = __importDefault(require("../dapp/artifacts/gasless_program.json"));
class GaslessTransaction {
    constructor(connection, wallet, dapp, gaslessType = types_1.GaslessTypes.POW) {
        this.connection = connection;
        this.wallet = wallet;
        this.dapp = dapp;
        this.transaction = new web3_js_1.Transaction();
        this.signers = [];
        this.gaslessType = gaslessType;
        const provider = new anchor_1.AnchorProvider(connection, wallet, { commitment: "confirmed" });
        const gaslessProgram = new anchor_1.Program(gasless_program_json_1.default, new web3_js_1.PublicKey("GasP6kcNpTdXA1M7ENyh5kCEvtHPgy71Habxe62gqHqH"), // TODO: get from gasless service
        provider);
        this.gaslessProgram = gaslessProgram;
    }
    static fromTransactionBuilder(connection, wallet, compressIx, dappUtil) {
        const { instructions, cleanupInstructions, signers } = compressIx;
        const _this = new GaslessTransaction(connection, wallet, dappUtil);
        _this.addSigners(signers);
        _this.addInstructions(instructions);
        _this.addInstructions(cleanupInstructions);
        return _this;
    }
    addSigners(signers) {
        this.signers = signers;
        return this;
    }
    addInstructions(ixs) {
        for (let i = 0; i < ixs.length; i++) {
            const ix = ixs[i];
            this.transaction.add(ix);
        }
        return this;
    }
    setGaslessType(gaslessType) {
        this.gaslessType = gaslessType;
        return this;
    }
    getPuzzleAndEstimateTime(reCaptchaToken) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.gaslessType !== types_1.GaslessTypes.POW) {
                throw Error(`${this.gaslessType} not supported`);
            }
            const puzzle = yield (0, gasless_1.getPuzzle)(this.connection, this.wallet.publicKey, reCaptchaToken);
            return { puzzle, estHandlingTime: pow_1.POWPuzzle.estHandlingTime(puzzle) };
        });
    }
    solveAndSubmit(puzzle) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.gaslessType !== types_1.GaslessTypes.POW) {
                throw Error(`${this.gaslessType} not supported`);
            }
            const { feePayer } = yield (0, gasless_1.getGaslessInfo)(this.connection);
            const txid = yield this._solveAndSubmitPuzzle(feePayer, puzzle);
            return txid;
        });
    }
    simulateGaslessTx(includeAccounts = []) {
        return __awaiter(this, void 0, void 0, function* () {
            const transaction = yield this.build();
            const simulated = yield this.connection.simulateTransaction(transaction, undefined, includeAccounts);
            return simulated.value;
        });
    }
    simulateRawTransaction(transaction, includeAccounts = []) {
        return __awaiter(this, void 0, void 0, function* () {
            const simulated = yield this.connection.simulateTransaction(transaction, undefined, includeAccounts);
            return simulated.value;
        });
    }
    build(sign = false) {
        return __awaiter(this, void 0, void 0, function* () {
            // Automatically detect the dapp type
            if (this.dapp.hasDappInstruction(this.transaction)) {
                this.gaslessType = types_1.GaslessTypes.Dapp;
            }
            const { feePayer } = yield (0, gasless_1.getGaslessInfo)(this.connection);
            if (this.gaslessType === types_1.GaslessTypes.Dapp) {
                this.transaction = yield this.dapp.build(this.transaction, this.wallet.publicKey, feePayer);
            }
            // pay for initializing mint account fee if needed
            this.transaction = helpers_1.TokenUtil.replaceFundingAccountOfCreateMintAccountIx(this.transaction, feePayer);
            // pay for initializing token account fee if needed
            this.transaction = helpers_1.TokenUtil.replaceFundingAccountOfCreateATAIx(this.transaction, feePayer);
            this.transaction.feePayer = feePayer;
            this.transaction.recentBlockhash = (yield this.connection.getRecentBlockhash()).blockhash;
            // Fund init accounts
            const { transaction, neededInits } = yield this._fundInitAccounts(this.transaction, feePayer);
            if (neededInits.length > 0) {
                // START: add verifyRentAccounts
                const fundAmounts = neededInits.map((v) => new anchor_1.BN(v.lamports));
                const accounts = neededInits.map((v) => {
                    return { pubkey: v.newAccount, isWritable: false, isSigner: false };
                });
                const ix = yield this.gaslessProgram.methods
                    .verifyRentAccounts(fundAmounts, "")
                    .accounts({
                    authority: feePayer,
                    rent: web3_js_1.SYSVAR_RENT_PUBKEY,
                })
                    .remainingAccounts(accounts)
                    .instruction();
                transaction.add(ix);
                this.transaction = transaction;
                // END: add verifyRentAccounts
            }
            if (sign) {
                for (let i = 0; i < this.signers.length; i++) {
                    const s = this.signers[i];
                    this.transaction.sign(s);
                }
                this.transaction = yield this.wallet.signTransaction(this.transaction);
            }
            return this.transaction;
        });
    }
    buildAndExecute(reCaptchaToken) {
        return __awaiter(this, void 0, void 0, function* () {
            // Automatically detect the dapp type
            if (this.dapp.hasDappInstruction(this.transaction)) {
                this.gaslessType = types_1.GaslessTypes.Dapp;
            }
            const { feePayer } = yield (0, gasless_1.getGaslessInfo)(this.connection);
            if (this.gaslessType === types_1.GaslessTypes.Dapp) {
                // this.transaction = await this.dapp.build(this.transaction, this.wallet.publicKey, feePayer);
                // // pay for initializing mint account fee if needed
                // this.transaction = TokenUtil.replaceFundingAccountOfCreateMintAccountIx(
                //   this.transaction,
                //   feePayer
                // );
                // // pay for initializing token account fee if needed
                // this.transaction = TokenUtil.replaceFundingAccountOfCreateATAIx(this.transaction, feePayer);
                // // TODO: add verifyRentAccounts
                // this.transaction.feePayer = feePayer;
                // this.transaction.recentBlockhash = (await this.connection.getRecentBlockhash()).blockhash;
                // for (let i = 0; i < this.signers.length; i++) {
                //   const s = this.signers[i];
                //   this.transaction.sign(s);
                // }
                // this.transaction = await this.wallet.signTransaction(this.transaction);
                const transaction = yield this.build(true);
                const txid = yield (0, gasless_1.sendToGasless)(this.connection, transaction, this.gaslessType, reCaptchaToken);
                return txid;
            }
            else if (this.gaslessType === types_1.GaslessTypes.POW) {
                const puzzle = yield (0, gasless_1.getPuzzle)(this.connection, this.wallet.publicKey, reCaptchaToken);
                const txid = yield this._solveAndSubmitPuzzle(feePayer, puzzle);
                return txid;
            }
            else {
                throw Error(`${this.gaslessType} not supported`);
            }
        });
    }
    solveAndSubmitPuzzle(feePayer, puzzle) {
        return __awaiter(this, void 0, void 0, function* () {
            return this._solveAndSubmitPuzzle(feePayer, puzzle);
        });
    }
    _solveAndSubmitPuzzle(feePayer, puzzle) {
        return __awaiter(this, void 0, void 0, function* () {
            const solution = yield pow_1.POWPuzzle.solveAsync(pow_1.Question.fromObject(puzzle.question));
            const rawSolution = Object.assign({ address: this.wallet.publicKey.toBase58(), solution: solution.toString(16) }, puzzle);
            // check if we can submit solution at this point (now >= puzzle.allowedSubmissionAt)
            const now = Math.floor(Date.now() / 1000);
            if (puzzle.allowedSubmissionAt && now < puzzle.allowedSubmissionAt) {
                yield (0, helpers_1.sleep)((puzzle.allowedSubmissionAt - now) * 1000);
            }
            // build transaction
            const transaction = yield this.build(true);
            // // pay for initializing mint account fee if needed
            // this.transaction = TokenUtil.replaceFundingAccountOfCreateMintAccountIx(
            //   this.transaction,
            //   feePayer
            // );
            // // pay for initializing token account fee if needed
            // this.transaction = TokenUtil.replaceFundingAccountOfCreateATAIx(this.transaction, feePayer);
            // this.transaction.feePayer = feePayer;
            // this.transaction.recentBlockhash = (await this.connection.getRecentBlockhash()).blockhash;
            // for (let i = 0; i < this.signers.length; i++) {
            //   const s = this.signers[i];
            //   this.transaction.sign(s);
            // }
            // this.transaction = await this.wallet.signTransaction(this.transaction);
            const txid = yield (0, gasless_1.postSolution)(this.connection, rawSolution, transaction);
            return txid;
        });
    }
    _fundInitAccounts(transaction, gaslessFeePayer) {
        return __awaiter(this, void 0, void 0, function* () {
            const initAccounts = [];
            for (let ix of transaction.instructions) {
                const feePayer = ix.keys.find((k) => k.isSigner);
                const keys = ix.keys.map((value) => value.pubkey);
                const infos = yield this.connection.getMultipleAccountsInfo(keys);
                // In transaction, the first signer must be feePayer of DApps and ignore the gasless's feePayer.
                if (feePayer && feePayer.pubkey.toBase58() !== gaslessFeePayer.toBase58()) {
                    for (let i = 0; i < keys.length; i++) {
                        const accountMeta = ix.keys[i];
                        const info = infos[i];
                        if (!info) {
                            initAccounts.push({
                                newAccount: accountMeta.pubkey,
                                feePayer: feePayer.pubkey,
                                lamports: 0,
                            });
                        }
                    }
                }
            }
            if (initAccounts.length > 0) {
                const ixs = transaction.instructions;
                let feePayer = null;
                const grouped = groupByToMap(initAccounts, (v) => v.feePayer);
                if (grouped.size > 1) {
                    throw "Only support a single fee payer";
                }
                grouped.forEach((v, k) => {
                    feePayer = k;
                    ixs.unshift(web3_js_1.SystemProgram.transfer({
                        fromPubkey: gaslessFeePayer,
                        toPubkey: feePayer,
                        lamports: 1000000000, // 1 RENECs
                    }));
                });
                if (feePayer) {
                    transaction.instructions = ixs;
                    if (initAccounts.length > 0) {
                        const simulated = (yield this.connection.simulateTransaction(transaction, undefined, initAccounts.map((v) => v.newAccount))).value;
                        if (simulated.err) {
                            throw simulated.err;
                        }
                        if (simulated.accounts && simulated.accounts.length > 0) {
                            let totalLamports = 0;
                            for (let i = 0; i < initAccounts.length; i++) {
                                const account = simulated.accounts[i];
                                if (account) {
                                    const data = Buffer.from(account.data[0], "base64");
                                    const lamports = yield this.connection.getMinimumBalanceForRentExemption(data.byteLength);
                                    initAccounts[i].lamports = lamports;
                                    totalLamports += lamports;
                                }
                            }
                            const ixs = transaction.instructions;
                            ixs.shift();
                            ixs.unshift(web3_js_1.SystemProgram.transfer({
                                fromPubkey: gaslessFeePayer,
                                toPubkey: feePayer,
                                lamports: totalLamports, // 1 RENECs
                            }));
                            transaction.instructions = ixs;
                            const finalSimulated = (yield this.connection.simulateTransaction(transaction, undefined, initAccounts.map((v) => v.newAccount))).value;
                            if (finalSimulated.err) {
                                throw finalSimulated.err;
                            }
                        }
                    }
                }
            }
            return { transaction, neededInits: initAccounts };
        });
    }
    asyncBuildAndExecute(cb) {
        setTimeout(() => __awaiter(this, void 0, void 0, function* () {
            try {
                const txid = yield this.buildAndExecute();
                cb(null, txid);
            }
            catch (e) {
                cb(e, "");
            }
        }), 0);
    }
}
exports.GaslessTransaction = GaslessTransaction;
const groupByToMap = (array, predicate) => array.reduce((map, value, index, array) => {
    var _a, _b;
    const key = predicate(value, index, array);
    (_b = (_a = map.get(key)) === null || _a === void 0 ? void 0 : _a.push(value)) !== null && _b !== void 0 ? _b : map.set(key, [value]);
    return map;
}, new Map());
